from collections import deque
from multiprocessing import Queue, Event

from cml.domain.deconstruction import *
from cml.shared.logger import DeconstInfo
from cml.domain.data_source import DataSource
from cml.domain.knowledge import KnowledgeDatabase, RelativeFinder, TS_QUEUE
from cml.shared.settings import DeconstructionSettings


__all__ = (
    "Deconstructor",
)


def reset_resources(func):
    def wrapper(*args, **kwargs):
        for log_info in func(*args, **kwargs):
            yield log_info

        self = args[0]
        self.upgrade_event.clear()
        self.apply_event.clear()
        self.clear_event.clear()

    return wrapper


class Deconstructor:
    TIME_COLUMN = "T"
    SUBJECT_COLUMN = "Sigma"
    PURPOSE_COLUMN = "Z"

    def __init__(self, mode, knowledge_database, relative_finder, source,
                 settings):
        self.mode = mode
        self.source = source
        self.settings = settings
        self.relative_finder = relative_finder
        self.knowledge_database = knowledge_database
        self.reconstructor = None

        # Used if parallelization is on
        self.tsigma_process = None
        self.tsigma_inputs = None
        self.tsigma_outputs = None
        self.upgrade_event = None
        self.apply_event = None
        self.clear_event = None

    def start_background_tsigma(self):
        if not self.tsigma_process:
            self.tsigma_inputs = Queue()
            self.tsigma_outputs = Queue()
            self.upgrade_event = Event()
            self.apply_event = Event()
            self.clear_event = Event()
            self.tsigma_process = TSigmaDeconstructor(self.source,
                                                      self.knowledge_database,
                                                      self.settings,
                                                      self.tsigma_inputs,
                                                      self.tsigma_outputs,
                                                      self.upgrade_event,
                                                      self.apply_event,
                                                      self.clear_event)
            self.tsigma_process.daemon = True
            self.tsigma_process.start()

        return self.tsigma_process

    @reset_resources
    def deconstruct_parallel(self, tier, pragmatic, block):
        info_status = []

        decon_one = self.start_background_tsigma()
        relatives = self.relative_finder.find(("T", "Sigma"), pragmatic)
        self.tsigma_inputs.put(Start(tier=tier,
                                     relatives=relatives,
                                     pragmatic=pragmatic,
                                     block=block))

        # complete deconstruction
        deq_two = deque()
        relatives = list(self.relative_finder.find(tuple(("complete", )), pragmatic))
        decon_two = CompleteDeconstructor(tier,
                                          self.source,
                                          self.knowledge_database,
                                          self.settings,
                                          self.reconstructor,
                                          pragmatic,
                                          relatives,
                                          block,
                                          deq_two)
        decon_two.run()

        while deq_two:
            log_info = deq_two.pop()
            yield log_info
            info_status.append(log_info.status)
            if log_info.status and self.settings.deconst_mode == "minimal":
                self.tsigma_inputs.put(Clean())
                self.clear_event.wait()
                return

            self.tsigma_inputs.put(Upgrade(deletes=log_info.deleted))

        self.tsigma_inputs.put(Apply())
        self.apply_event.wait()
        while not self.tsigma_outputs.empty():
            log_info, tier, lb = self.tsigma_outputs.get(timeout=0)
            TS_QUEUE.append((tier, lb))
            yield log_info
            info_status.append(log_info)
            if log_info.status and self.settings.deconst_mode == "minimal":
                return

        deq_three = deque()
        relatives = list(self.relative_finder.find(("Sigma", "Z"), pragmatic))
        decon_three = ZSigmaDeconstructor(tier,
                                          self.source,
                                          self.knowledge_database,
                                          self.settings,
                                          self.reconstructor,
                                          pragmatic,
                                          relatives,
                                          block,
                                          deq_three)
        decon_three.run()
        while deq_three:
            log_info = deq_three.pop()
            yield log_info
            info_status.append(log_info)
            if log_info.status and self.settings.deconst_mode == "minimal":
                return

        if not info_status or not all(info_status):
            self.knowledge_database.insert(pragmatic)
            yield self.arrange_log(pragmatic)

    def deconstruct_serial(self, tier, pragmatic, block):
        changes_done_status = []

        relatives = self.relative_finder.find(tuple(("complete", )), pragmatic)
        deq_one = deque()
        decon_one = CompleteDeconstructor(tier,
                                          self.source,
                                          self.knowledge_database,
                                          self.settings,
                                          self.reconstructor,
                                          pragmatic,
                                          relatives,
                                          block,
                                          deq_one)
        decon_one.run()
        for log in deq_one:
            yield log
            changes_done_status.append(log.status)
            if log.status and self.settings.deconst_mode == "minimal":
                return
            elif pragmatic in log.deleted:
                return

        relatives = self.relative_finder.find(("T", "Sigma"), pragmatic)
        deq_two = deque()
        decon_two = TSigmaDeconstructor(self.source,
                                        self.knowledge_database,
                                        self.settings)
        decon_two.run_serial(tier, relatives, pragmatic, block, deq_two)
        for log, tier_, lb in deq_two:
            TS_QUEUE.append((tier_, lb))
            yield log
            changes_done_status.append(log.status)
            if log.status and self.settings.deconst_mode == "minimal":
                return
            elif pragmatic in log.deleted:
                return

        relatives = self.relative_finder.find(("Sigma", "Z"), pragmatic)
        deq_three = deque()
        decon_three = ZSigmaDeconstructor(tier,
                                          self.source,
                                          self.knowledge_database,
                                          self.settings,
                                          self.reconstructor,
                                          pragmatic,
                                          relatives,
                                          block,
                                          deq_three)
        decon_three.run()
        for log in deq_three:
            yield log
            self.clean_ts_queue(log.deleted)
            changes_done_status.append(log.status)
            if log.status and self.settings.deconst_mode == "minimal":
                return
            elif pragmatic in log.deleted:
                return

        if not changes_done_status or not all(changes_done_status):
            self.knowledge_database.insert(pragmatic)
            yield self.arrange_log(pragmatic)

    def clean_ts_queue(self, deleted):
        queue_copy = TS_QUEUE.copy()
        for tier, lb in queue_copy:
            for model in deleted:
                if model in lb.origin:
                    TS_QUEUE.remove((tier, lb))

    def arrange_log(self, model):
        log = DeconstInfo()
        log.pragmatic = model
        log.inserted.append(model)
        return log
