from contextlib import suppress

from cml.domain.deconstruction.abstract import AbstractDeconstructor
from cml.shared.errors import (
    NotEnoughFeaturesWarning,
    ReconstructionFailure)

__all__ = (
    "CompleteDeconstructor",
)


class CompleteDeconstructor(AbstractDeconstructor):
    RELATION = ("complete", )
    MIN_N_FEATURE = 2

    def __init__(self, *args):
        super().__init__(*args)

    def deconstruct(self, relative, log):
        relative_block = relative.trained_with(self.source, self.knowledge)
        if self.feature_constraint(relative):
            meta = self.model_fusion(relative)
            learnblock = self.block.same_features_fusion(relative_block)
            return self.reconstructing(meta, learnblock, relative, log)

        elif self.reliability_constraint(relative_block):
            meta = self.model_fusion(relative)
            learnblock = self.build_block_from_diff_features(relative_block)
            return self.reconstructing(meta, learnblock, self.pragmatic, log)

        else:
            modification = self.model_disposal(self.pragmatic, relative,
                                               self.block, relative_block)
            log.inserted = modification.inserted
            log.deleted = modification.deleted
            return log

    def feature_constraint(self, relative):
        first_block = self.pragmatic.trained_with(self.source, self.knowledge)
        second_block = relative.trained_with(self.source, self.knowledge)
        identicial_features = len(set(first_block.columns()).intersection(
            set(second_block.columns())))
        return identicial_features >= self.MIN_N_FEATURE

    def reconstructing(self, meta, learnblock, relative, log):
        try:
            reconstructed = self.reconstructor.reconstruct_serial(
                self.tier, learnblock, meta.sigma, meta)
            reconstructed.identifier = self.correct_reconstructed_identifier(self.tier)
            assert reconstructed.tier == self.tier
            log.reconstructed_info.append(reconstructed)
            self.knowledge.remove(relative)
            self.knowledge.insert(reconstructed)
            log.inserted.append(reconstructed)
            log.deleted.append(relative)
            log.status = True
            return log

        except (ReconstructionFailure, NotEnoughFeaturesWarning):
            return self.model_differentiation(self.tier, learnblock, relative,
                                              log)

    def model_differentiation(self, tier, block, relative_model, log):
        time_column = block.get_column("T")
        density = self.source.estimate_density(time_column)
        changes = self.source.remove_time_dense_relatives(density)
        clusters = self.source.cluster(block, changes)

        inserted = []
        for time_values in clusters:
            learnblock = block.new_block_from(time_values, time_column=True)
            with suppress(ReconstructionFailure, NotEnoughFeaturesWarning):
                reconstructed = self.reconstructor.reconstruct_serial(tier, learnblock)
                reconstructed.identifier = self.correct_reconstructed_identifier(self.tier)
                log.reconstructed_info.append(reconstructed)
                log.inserted.append(reconstructed)
                self.knowledge.insert(reconstructed)

        if not inserted:
            return log

        else:
            to_delete = self.knowledge.remove_dependent_models(relative_model)
            for model in to_delete:
                log.deleted.append(model)
                self.knowledge.remove(model)
            log.status = True
            return log

    def reliability_constraint(self, relative_block):
        overlapped = self.block.get_overlapping(relative_block)
        reliability = self.calculate_reliability(
            overlapped.get_column("Z"),
            overlapped.get_column("Z_other"))
        return reliability >= self.settings.min_reliability

    def build_block_from_diff_features(self, relative_block):
        features = set(self.block.n_features).difference(
            set(relative_block.n_features))
        features.add("T")
        return self.block.get_overlapping(relative_block,
                                          drop_not=tuple(features))

